# -*- coding: utf-8 -*-

from __future__ import (
    print_function, division, generators,
    absolute_import, unicode_literals
)

from datetime import datetime
import time

from nexusadspy import AppnexusClient
from nexusadspy.exceptions import NexusadspyAPIError


class AppnexusReport():
    def __init__(self, report_type, columns, timezone='CET', filters=None,
                 groups=None, start_date=None, end_date=None, report_interval=None,
                 advertiser_ids=None, publisher_ids=None,
                 credentials_path='.appnexus_auth.json',
                 max_retries=100, retry_seconds=2.):
        """
        AppNexus reporting class.

        :param report_type: str
        :param columns: list
        :param timezone: str
        :param filters: list
        :param groups: list
        :param start_date: str
        :param end_date: str
        :param report_interval: str
        :param advertiser_ids: int
        :param publisher_ids: int
        :param credentials_path: str
        :param max_retries: int
        :param retry_seconds: float
        :return:
        """

        if not isinstance(columns, list):
            raise ValueError('"columns" is expected as a list, you '
                             'provided "{}"'.format(columns))

        self.report_type = report_type
        self.columns = columns
        self.timezone = timezone
        self.filters = filters or []
        self.groups = groups or []
        self.start_date = self._format_date(start_date)
        self.end_date = self._format_date(end_date)
        self.report_interval = report_interval
        self.advertiser_ids = advertiser_ids or []
        self.publisher_ids = publisher_ids or []
        self.credentials_path = credentials_path
        self.max_retries = max_retries
        self.retry_seconds = retry_seconds

        self.request = self._build_request()
        self.endpoint = self._build_endpoint()

        self._handle_network_user_request()

    def get(self, format_='json'):
        """
        Trigger and download the report.

        :param format_: optional, Specify 'pandas' to get report as a DataFrame.
        :return:
        """
        client = AppnexusClient(self.credentials_path)
        response = self._post_request(client)
        report_id = response['report_id']

        report = self._get_report(client, report_id)

        if format_ == 'pandas':
            report = self._convert_to_dataframe(report)

        return report

    def _post_request(self, client):
        response = client.request(self.endpoint, 'POST', data=self.request)

        return response[0]

    def _get_report(self, client, report_id):
        self._poll_and_wait(client, report_id)  # block until report ready
        report = self._download_report(client, report_id)

        return report

    def _poll_and_wait(self, client, report_id):
        response = {}
        for retry in range(self.max_retries):
            data = {'id': report_id}
            response = client.request(self.endpoint, 'GET', data=data)[0]
            execution_status = self._get_report_execution_status(response, report_id)
            if execution_status == 'ready':
                response = client.request(self.endpoint, 'GET', data=data, get_field='report')
                break
            else:
                time.sleep(self.retry_seconds**retry)

        if retry == self.max_retries - 1:
            raise NexusadspyAPIError('Report with ID "{}" not ready. '
                                     'Last response was "{}".'.format(report_id,
                                                                      response))

    @staticmethod
    def _get_report_execution_status(response, report_id):
        try:
            return response['execution_status']
        except KeyError:
            report = AppnexusReport._get_report_dict(response['reports'], report_id)
            return report['execution_status']

    @staticmethod
    def _get_report_dict(reports, report_id):
        for report in reports:
            if report['id'] == report_id:
                return report

    @staticmethod
    def _convert_to_dataframe(report):
        import pandas as pd

        return pd.DataFrame(report)

    @staticmethod
    def _download_report(client, report_id):
        report = client.request('report-download', 'GET', get_field='report',
                                params={'id': report_id})

        return report

    @staticmethod
    def _format_date(date_string):
        try:
            datetime.strptime(date_string, "%Y-%m-%d %H:%M:%S")
        except ValueError:
            try:
                date_string = datetime.strptime(date_string, "%Y-%m-%d")
                date_string = date_string.strftime("%Y-%m-%d %H:%M:%S")
            except:
                raise ValueError('Expected data format is YYYY-MM-DD. You '
                                 'provided "{}".'.format(date_string))
        except TypeError:
            pass

        return date_string

    def _build_request(self):
        r = self._get_request_skeleton()
        r = self._add_request_date(r)
        r = self._add_request_filters(r)
        r = self._add_request_groups(r)

        return r

    def _get_request_skeleton(self):
        r = {
            'report': {
                'special_pixel_reporting': False,
                'report_type': self.report_type,
                'timezone': self.timezone,
                'columns': self.columns
            }
        }

        return r

    def _add_request_date(self, r):
        if self.start_date and self.end_date:
            r['report']['start_date'] = self.start_date
            r['report']['end_date'] = self.end_date
        elif self.report_interval:
            r['report_interval'] = self.report_interval
        else:
            r['report_interval'] = 'lifetime'

        return r

    def _add_request_filters(self, r):
        r['report']['filters'] = self.filters

        return r

    def _add_request_groups(self, r):
        r['report']['groups'] = self.groups

        return r

    @staticmethod
    def _build_endpoint():
        return 'report'

    def _handle_network_user_request(self):
        if self.advertiser_ids and self.publisher_ids:
            raise ValueError('As a network user you can request '
                             'advertiser-level and publisher-level reports '
                             'but not both at the same time. You provided '
                             '"advertiser_ids={}" and "publisher_ids={}".'.format(self.advertiser_ids,
                                                                                  self.publisher_ids))

        if self.advertiser_ids:
            self._amend_request('advertiser_id', self.advertiser_ids)
            self._amend_endpoint('advertiser_id', self.advertiser_ids)
        elif self.publisher_ids:
            self._amend_request('publisher_id', self.publisher_ids)
            self._amend_endpoint('publisher_id', self.publisher_ids)

    def _amend_request(self, identifier, items):
        if identifier not in self.request['report']['columns']:
            self.request['report']['columns'].append(identifier)

        self.request['report']['filters'].append({identifier: items})

    def _amend_endpoint(self, identifier, items):
        pass



# -*- coding: utf-8 -*-

from __future__ import (
    print_function, division, generators,
    absolute_import, unicode_literals
)

from nexusadspy.client import AppnexusClient  # NOQA
from nexusadspy.report import AppnexusReport  # NOQA
from nexusadspy.segment import AppnexusSegmentsUploader  # NOQA




from __future__ import (
    print_function, division, generators,
    absolute_import, unicode_literals
)

from io import BytesIO
from gzip import GzipFile
import time
import logging

from nexusadspy.client import AppnexusClient


class AppnexusSegmentsUploader:

    BATCH_UPLOAD_ANDROID_SPECIFIER = '8'
    BATCH_UPLOAD_IOS_SPECIFIER = '3'

    def __init__(self, users_list, segment_code, separators, member_id,
                 credentials_path='.appnexus_auth.json'):
        """
        Batch-upload API wrapper for AppNexus.
        :param users_list: list, List of dictionaries representing AppNexus users. Every member should have fields
            - uid: AppNexus user ID. AAID/IDFS in case of mobile.
            - timestamp: POSIX timestamp when user entered the segment.
            - expiration (optional): Expiration timestamp for the user. A POSIX timestamp. Defaults to 0.
            - value (optional): Numerical value for the segment. Defaults to 0.
            - mobile_os (optional): OS used by the user. Considered internally by AppNexus to be desktop if absent.
        :param segment_code: str, Segment code to add users to.
        :param separators: list, List of five field separators. As documented in
        https://wiki.appnexus.com/display/api/Batch+Segment+Service+-+File+Format#BatchSegmentService-FileFormat-Separators
        :param member_id: str, Member ID for AppNexus account.
        :param credentials_path: str (optional), Credentials path for AppnexusClient. Defaults to '.appnexus_auth.json'.
        :return:
        """
        self._credentials_path = credentials_path
        self._users_list = users_list
        self._segment_code = segment_code  # Appnexus bug: Segment upload batch API does not work with segment IDs
        self._separators = separators
        self._member_id = member_id
        self._logger = logging.getLogger('nexusadspy.segment')

    def upload(self, polling_duration_sec=2, max_retries=10):
        """
        Initiate segment upload task
        :param polling_duration_sec: int (optional), Time to sleep while polling for status. Defaults to 2.
        :param max_retries: int (optional), Max number of polling retries to be done. Defaults to 10.
        :return: tuple, Tuple with two values, number of valid users and invalid users.
        """
        valid_user_count = invalid_user_count = 0
        api_client = AppnexusClient(self._credentials_path)
        job_id, upload_url = self._initialize_job(api_client)
        self._upload_batch_to_url(api_client, upload_url)
        for attempt in range(max_retries):
            time.sleep(polling_duration_sec)
            job_status = self._get_job_status_response(api_client, job_id)
            if job_status[0].get('phase') == 'completed':
                valid_user_count = job_status[0].get('num_valid_user')
                invalid_user_count = job_status[0].get('num_invalid_user')
                break
        return valid_user_count, invalid_user_count

    def _initialize_job(self, api_client):
        service_endpoint = 'batch-segment?member_id={}'.format(self._member_id)
        response = api_client.request(service_endpoint, 'POST')
        job_id = response[0]['batch_segment_upload_job']['job_id']
        upload_url = response[0]['batch_segment_upload_job']['upload_url']
        return job_id, upload_url

    def _upload_batch_to_url(self, api_client, upload_url):
        upload_buffer = self._get_buffer_for_upload()
        headers = {'Content-Type': 'application/octet-stream'}
        api_client.request(upload_url, 'POST', data=upload_buffer.read(), prepend_endpoint=False, headers=headers)

    def _get_job_status_response(self, api_client, job_id):
        status_endpoint = 'batch-segment?member_id={}&job_id={}'.format(self._member_id, job_id)
        headers = {'Content-Type': 'application/octet-stream'}
        return api_client.request(status_endpoint, 'GET', headers=headers)

    def _get_buffer_for_upload(self):
        upload_string = '\n'.join(self._get_upload_string_for_user(user) for user in self._users_list)
        self._logger.debug("Attempting to upload \n" + upload_string)
        compressed_buffer = BytesIO()
        with GzipFile(fileobj=compressed_buffer, mode='wb') as compressor:
            compressor.write(upload_string.encode('UTF-8'))
        compressed_buffer.seek(0)
        return compressed_buffer

    def _get_upload_string_for_user(self, user):
        upload_string = str(user['uid']) + self._separators[0]
        upload_string += str(self._segment_code) + self._separators[2]
        upload_string += str(user.get('expiration', 0)) + self._separators[2]
        upload_string += str(user['timestamp']) + self._separators[2]
        upload_string += str(user.get('value', 0))
        user_mobile_os = user.get('mobile_os')
        if user_mobile_os is not None:
            if user_mobile_os.lower() == 'android':
                # Appnexus bug: AAID should be uploaded as both IDFA and AAID. Otherwise it cannot be used in mopub.
                upload_string = upload_string + self._separators[4] + self.BATCH_UPLOAD_ANDROID_SPECIFIER +\
                    "\n" + upload_string + self._separators[4] + self.BATCH_UPLOAD_IOS_SPECIFIER
            elif user_mobile_os.lower() == 'ios':
                upload_string = upload_string + self._separators[4] + self.BATCH_UPLOAD_IOS_SPECIFIER
        return upload_string



# -*- coding: utf-8 -*-

from __future__ import (
    print_function, division, generators,
    absolute_import, unicode_literals
)

from collections import defaultdict
import os
import time
import json
import logging

try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

try:
    import FileNotFoundError
except ImportError as err:
    FileNotFoundError = IOError

from nexusadspy.exceptions import NexusadspyAPIError, NexusadspyConfigurationError

import requests


logging.basicConfig(level=logging.INFO)


class AppnexusClient:

    def __init__(self, path, endpoint='https://api.appnexus.com'):
        self.path = path
        self.endpoint = endpoint
        self._session = None
        self.logger = logging.getLogger('AppnexusClient')

    def request(self, service, method, params=None, data=None, headers=None,
                get_field=None, prepend_endpoint=True):
        """
        Sends a request to the Appnexus API. Handles authentication, paging, and throttling.

        :param service: str, One of the services Appnexus services (https://wiki.appnexus.com/display/api/API+Services).
        :param method: str, HTTP method to be used. One of 'GET', 'POST', 'PUT', or 'DELETE'.
        :param params: dict (optional), Any data to be sent in URL as parameters.
        :param data: dict (optional), Any data to be sent in the request.
        :param headers: dict (optional), Any HTTP headers to be sent in the request.
        :return: list, List of response dictionaries.
        """
        method = method.lower()

        params = params or {}
        data = data or {}

        if method not in ['get', 'post', 'put', 'delete']:
            raise ValueError(
                'Argument "method" must be one of '
                '["get", "post", "put", "delete"]. '
                'You supplied: "{}".'.format(method)
            )

        url = urljoin(base=self.endpoint, url=service) if prepend_endpoint else service

        if method == 'get':
            res_code, res = self._do_paged_get(url, method, params=params,
                                               data=data, headers=headers,
                                               get_field=get_field)
        else:
            res_code, res = self._do_authenticated_request(url, method,
                                                           params=params,
                                                           data=data,
                                                           headers=headers)

        self._check_response(res_code, res)

        if not isinstance(res, list):
            res = [res]

        return res

    @property
    def session(self):
        self._session = self._session or requests.Session()
        return self._session

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()

    def _do_paged_get(self, url, method, params=None, data=None, headers=None,
                      start_element=None, batch_size=None, max_items=None,
                      get_field=None):
        res = defaultdict(list)

        if start_element is None:
            start_element = 0
        if batch_size is None:
            batch_size = 100

        while True:
            data.update({'start_element': start_element,
                         'batch_size': batch_size})

            r_code, r = self._do_authenticated_request(url, method, params=params,
                                                       data=data, headers=headers,
                                                       get_field=get_field)

            output_term = get_field or r['dbg_info']['output_term']
            output = r.get(output_term, r)

            if isinstance(output, list):
                res[output_term] += output  # assume list of dictionaries
            else:
                if not isinstance(output, dict):
                    res[output_term].append({output_term: output})
                else:
                    res[output_term].append(output)

            start_element += batch_size

            count = int(r.get('count', 0) or 0)
            if len(res[output_term]) >= count:
                break

            if max_items is not None and len(res) >= max_items:
                break

        res = res[output_term]

        return r_code, res

    def _do_throttled_request(self, url, method, params=None, data=None, headers=None,
                              sec_sleep=2., max_failures=100,
                              get_field=None):

        if isinstance(data, dict):
            data = json.dumps(data)
        no_fail = 0
        while True:
            r = self.session.request(method, url, params=params, data=data, headers=headers)
            r_code = r.status_code

            try:
                headers = r.headers
                r = r.json()['response']
                r['headers'] = headers
            except (KeyError, ValueError):
                if len(r.content) > 0:
                    r = self._convert_csv_to_dict(r.content, get_field)
                else:
                    self._check_response(r_code, {})

            if no_fail < max_failures and r.get('error_code', '') == 'RATE_EXCEEDED':
                no_fail += 1
                time.sleep(sec_sleep ** no_fail)
                continue

            return r_code, r

    @staticmethod
    def _convert_csv_to_dict(csv_bytestr, field):
        s = csv_bytestr.decode('utf-8')
        headings, rows = s.split('\r\n')[0], s.split('\r\n')[1:]
        headings = [h.strip() for h in headings.split(',')]
        rows = (r for r in rows if len(r) > 0)
        rows = (r.split(',') for r in rows)
        rows = [[el.strip() for el in r] for r in rows]

        res = [{h: v for h, v in zip(headings, row)} for row in rows]

        return {field: res}

    def _do_authenticated_request(self, url, method, params=None, data=None,
                                  headers=None, get_field=None):
        headers = headers or {}
        headers.update({'Authorization': self._get_auth_token()})

        while True:
            r_code, r = self._do_throttled_request(url, method, params=params,
                                                   data=data, headers=headers,
                                                   get_field=get_field)

            if r.get('error_id', '') == 'NOAUTH':
                headers.update({'Authorization': self._get_auth_token(overwrite=True)})
                continue  # retry with new authorization token

            return r_code, r

    def _get_auth_token(self, overwrite=False):
        if overwrite:
            os.rename(self.path, self.path + '_backup')

        try:
            token = self._get_cached_auth_token()
        except FileNotFoundError:
            token = self._get_new_auth_token()
            self._cache_auth_token(token)
        return token

    def _cache_auth_token(self, token):
        with open(self.path, 'w') as f:
            json.dump({'token': token}, f)

    def _get_cached_auth_token(self):
        with open(self.path, 'r') as f:
            auth = json.load(f)
            return auth['token']

    def _get_new_auth_token(self):
        try:
            username = os.environ['USERNAME_NEXUSADSPY']
            password = os.environ['PASSWORD_NEXUSADSPY']
        except KeyError as e:
            raise NexusadspyConfigurationError(
                'Set environment variables "USERNAME_NEXUSADSPY" and '
                '"PASSWORD_NEXUSADSPY" appropriately. '
                'You failed to set: "{}".'.format(e.args[0])
            )

        data = {'auth': {'username': username, 'password': password}}
        headers = {'Content-type': 'application/json; charset=UTF-8'}
        url = urljoin(base=self.endpoint, url='auth')

        r_code, r = self._do_throttled_request(url, 'post', data=data, headers=headers)
        self._check_response(r_code, r)

        token = r['token']

        return token

    def _check_response(self, response_code, response):
        if not isinstance(response, list):
            response = [response]

        for res in response:
            if res.get('error_id') is not None or response_code != 200:
                raise NexusadspyAPIError('Response status code: "{}"'.format(response_code),
                                         res.get('error_id'),
                                         res.get('error'),
                                         res.get('error_description'))



# !/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name='nexusadspy',
    version='0.3.1',
    description='Thin wrapper around AppNexus API',
    author='Daniel Olel, Georg Walther',
    author_email='daniel.olel@rocket-internet.com, '
                 'georg.walther@markovian.com',
    url='https://github.com/markovianhq/nexusadspy',
    download_url='https://github.com/markovianhq/nexusadspy/tarball/master',
    packages=['nexusadspy'],
    package_dir={'nexusadspy': 'nexusadspy'},
    package_data={},
    include_package_data=False,
    zip_safe=False,
    keywords='nexuadspy appnexus api python',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ]
)



# -*- coding: utf-8 -*-

from __future__ import (
    print_function, division, generators,
    absolute_import, unicode_literals
)


class NexusadspyError(Exception):
    pass


class NexusadspyAPIError(NexusadspyError):
    pass


class NexusadspyConfigurationError(NexusadspyError):
    pass
